import os
from math import floor
from typing import Optional

import platformdirs
import pygame

from src.image_handler import ImageHandler


class Editor:
    WIDTH: int = 800
    HEIGHT: int = 600
    MAP_SIZE: int = 1024
    MAPS_DIR: str = os.path.join(platformdirs.user_cache_dir(appname="lycacraft-paths", appauthor="Lycacraft"), "maps")
    ZOOMS: tuple[float] = (0.25, 0.5, 1, 2, 4)
    CROSSHAIR_SIZE: int = 10

    def __init__(self):
        pygame.init()
        self.width: int = self.WIDTH
        self.height: int = self.HEIGHT
        self.win: pygame.Surface = pygame.display.set_mode([self.width, self.height], pygame.RESIZABLE)
        pygame.display.set_caption("Lycacraft Map Editor")
        self.center: list[int] = [0, 0]
        self.zoom_i: int = 2
        self.zoom: float = self.ZOOMS[self.zoom_i]
        self.running: bool = False
        self.image_handler: ImageHandler = ImageHandler(self.MAPS_DIR, self.MAP_SIZE)
        self.clock: pygame.time.Clock = pygame.time.Clock()
        self.drag_pos: Optional[tuple[int, int]] = None
        self.font: pygame.font.Font = pygame.font.SysFont("Ubuntu", 20)
        self.zooms_texts: list[pygame.Surface] = list(map(
            lambda z: self.font.render(str(z), True, (255, 255, 255)),
            self.ZOOMS
        ))

    def mainloop(self) -> None:
        self.running = True
        while self.running:
            pygame.display.set_caption(f"Lycacraft Map Editor - {self.clock.get_fps():.2f}fps")
            self.process_events()
            self.render()
            self.clock.tick(30)

    def process_events(self) -> None:
        events = pygame.event.get()

        for event in events:
            if event.type == pygame.QUIT:
                self.running = False
            elif event.type == pygame.WINDOWRESIZED:
                self.width = event.x
                self.height = event.y
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    self.running = False
                elif event.key == pygame.K_PAGEUP:
                    self.zoom_in()
                elif event.key == pygame.K_PAGEDOWN:
                    self.zoom_out()
            elif event.type == pygame.MOUSEBUTTONDOWN:
                if event.button == 2:
                    self.drag_pos = event.pos
            elif event.type == pygame.MOUSEBUTTONUP:
                if event.button == 2:
                    self.drag_pos = None

        keys = pygame.key.get_pressed()
        if keys[pygame.K_LEFT]:
            self.center[0] -= 4 / self.zoom
        if keys[pygame.K_RIGHT]:
            self.center[0] += 4 / self.zoom
        if keys[pygame.K_UP]:
            self.center[1] -= 4 / self.zoom
        if keys[pygame.K_DOWN]:
            self.center[1] += 4 / self.zoom

        mbtns = pygame.mouse.get_pressed()
        mpos = pygame.mouse.get_pos()
        if mbtns[1]:
            dx = mpos[0] - self.drag_pos[0]
            dy = mpos[1] - self.drag_pos[1]
            self.center[0] -= dx / self.zoom
            self.center[1] -= dy / self.zoom
            self.drag_pos = mpos

    def render(self) -> None:
        self.win.fill((0, 0, 0))
        off_x = (self.center[0] * self.zoom) % self.MAP_SIZE
        off_y = (self.center[1] * self.zoom) % self.MAP_SIZE

        w2 = self.width / 2
        h2 = self.height / 2

        # In game top-left / bottom-right corners
        x0 = floor(self.center[0] - w2 / self.zoom)
        y0 = floor(self.center[1] - h2 / self.zoom)
        x1 = floor(self.center[0] + w2 / self.zoom)
        y1 = floor(self.center[1] + h2 / self.zoom)

        # Top-left / bottom-right maps
        mx0 = floor(x0 * self.zoom / self.MAP_SIZE)
        my0 = floor(y0 * self.zoom / self.MAP_SIZE)
        mx1 = floor(x1 * self.zoom / self.MAP_SIZE)
        my1 = floor(y1 * self.zoom / self.MAP_SIZE)

        h_maps = mx1 - mx0 + 1
        v_maps = my1 - my0 + 1
        cx = floor(self.center[0] * self.zoom / self.MAP_SIZE)
        cy = floor(self.center[1] * self.zoom / self.MAP_SIZE)
        ox = w2 + (mx0 - cx) * self.MAP_SIZE - off_x
        oy = h2 + (my0 - cy) * self.MAP_SIZE - off_y

        for y in range(v_maps):
            for x in range(h_maps):
                map_image = self.image_handler.get_for_pos(mx0 + x, my0 + y, self.zoom)
                if map_image is None:
                    continue
                self.win.blit(map_image, [
                    ox + x * self.MAP_SIZE,
                    oy + y * self.MAP_SIZE
                ])

        pygame.draw.line(self.win, (150, 150, 150), [w2 - self.CROSSHAIR_SIZE, h2], [w2 + self.CROSSHAIR_SIZE, h2])
        pygame.draw.line(self.win, (150, 150, 150), [w2, h2 - self.CROSSHAIR_SIZE], [w2, h2 + self.CROSSHAIR_SIZE])
        self.render_zoom_slider()

        mpos = pygame.mouse.get_pos()
        world_x, world_z = self.screen_to_world(*mpos)
        mouse_txt = self.font.render(f"x: {world_x} / z: {world_z}", True, (255, 255, 255))
        pygame.draw.rect(self.win, (80, 80, 80), [0, 0, mouse_txt.get_width() + 10, mouse_txt.get_height() + 10])
        self.win.blit(mouse_txt, [5, 5])

        pygame.display.flip()

    def render_zoom_slider(self) -> None:
        zoom_height = self.height * 0.2
        zoom_h_margin = self.width * 0.02
        zoom_v_margin = self.height * 0.05
        zoom_x = self.width - zoom_h_margin
        zoom_y = self.height - zoom_v_margin - zoom_height
        zoom_space = zoom_height / 4
        zoom_r = zoom_space / 4
        zoom_width = max(s.get_width() for s in self.zooms_texts) + 2 * zoom_r + 5
        pygame.draw.rect(self.win, (80, 80, 80), [
            zoom_x + zoom_r - zoom_width - 5,
            zoom_y - zoom_r - 5,
            zoom_width + 10,
            zoom_height + 2 * zoom_r + 10
        ])
        pygame.draw.line(self.win, (255, 255, 255), [zoom_x, zoom_y], [zoom_x, zoom_y + zoom_height])
        for i, txt in enumerate(self.zooms_texts):
            y = zoom_y + zoom_height - i * zoom_space
            col = (255, 0, 0) if i == self.zoom_i else (255, 255, 255)
            pygame.draw.circle(self.win, col, [zoom_x, y], zoom_r)
            self.win.blit(txt, [zoom_x - txt.get_width() - zoom_r - 5, y - txt.get_height() / 2])

    def set_zoom(self, zoom_i: int) -> None:
        self.zoom_i = max(0, min(len(self.ZOOMS) - 1, zoom_i))
        self.zoom = self.ZOOMS[self.zoom_i]

    def zoom_in(self) -> None:
        self.set_zoom(self.zoom_i + 1)

    def zoom_out(self) -> None:
        self.set_zoom(self.zoom_i - 1)

    def screen_to_world(self, x: int, y: int) -> tuple[int, int]:
        w2 = self.width / 2
        h2 = self.height / 2
        world_x = floor((x - w2) / self.zoom + self.center[0])
        world_z = floor((y - h2) / self.zoom + self.center[1])

        return int(world_x), int(world_z)
